'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var view = require('@codemirror/view');
var state = require('@codemirror/state');

// TODO: custom style
const interactField = state.StateField.define({
    create: () => null,
    update: (value, tr) => {
        for (const e of tr.effects) {
            if (e.is(setInteract)) {
                return e.value;
            }
        }
        if (!value) {
            return null;
        }
        if (!tr.changes.empty) {
            const newPos = tr.changes.mapPos(value.pos, -1, state.MapMode.TrackDel);
            const newEnd = tr.changes.mapPos(value.pos + value.text.length, -1, state.MapMode.TrackDel);
            if (newPos === null || newEnd === null) {
                return null;
            }
            // if the text doesn't match anymore, we'll just return null
            // rather than checking if the rule matches again
            if (tr.newDoc.sliceString(newPos, newEnd) !== value.text) {
                return null;
            }
            return Object.assign(Object.assign({}, value), { pos: newPos });
        }
        return value;
    },
    provide: (field) => [
        view.EditorView.decorations.from(field, (target) => {
            if (!target) {
                return view.Decoration.none;
            }
            const from = target.pos;
            const to = target.pos + target.text.length;
            return view.Decoration.set(mark.range(from, to));
        }),
        view.EditorView.contentAttributes.from(field, (target) => {
            if (!target || !target.rule.cursor) {
                return { style: '' };
            }
            return { style: `cursor: ${target.rule.cursor}` };
        }),
    ]
});
const setInteract = state.StateEffect.define();
const mark = view.Decoration.mark({ class: 'cm-interact' });
const interactTheme = view.EditorView.theme({
    '.cm-interact': {
        background: 'rgba(128, 128, 255, 0.2)',
        borderRadius: '4px'
    },
});
/**
 * A rule that defines a type of value and its interaction.
 *
 * @example
 * ```
 * // a number dragger
 * interactRule.of({
 *     // the regexp matching the value
 *     regexp: /-?\b\d+\.?\d*\b/g,
 *     // set cursor to 'ew-resize'on hover
 *     cursor: 'ew-resize'
 *     // change number value based on mouse X movement on drag
 *     onDrag: (text, setText, e) => {
 *         const newVal = Number(text) + e.movementX;
 *         if (isNaN(newVal)) return;
 *         setText(newVal.toString());
 *     },
 * })
 * ```
 */
const interactRule = state.Facet.define();
const interactModKey = state.Facet.define({
    combine: (values) => values[values.length - 1],
});
const interactViewPlugin = view.ViewPlugin.define((view) => ({
    target: null,
    dragging: false,
    mouseX: 0,
    mouseY: 0,
    // Get current match under cursor from all rules
    getMatch() {
        const rules = view.state.facet(interactRule);
        const pos = view.posAtCoords({ x: this.mouseX, y: this.mouseY });
        if (!pos)
            return null;
        const line = view.state.doc.lineAt(pos);
        const lpos = pos - line.from;
        let match = null;
        for (const rule of rules) {
            for (const m of line.text.matchAll(rule.regexp)) {
                if (m.index === undefined)
                    continue;
                const text = m[0];
                if (!text)
                    continue;
                const start = m.index;
                const end = m.index + text.length;
                if (lpos < start || lpos > end)
                    continue;
                // If there are overlap matches from different rules, use the smaller one
                if (!match || text.length < match.text.length) {
                    match = {
                        rule: rule,
                        pos: line.from + start,
                        text: text,
                    };
                }
            }
        }
        return match;
    },
    updateText(target) {
        return (text) => {
            view.dispatch({
                effects: setInteract.of(Object.assign(Object.assign({}, target), { text })),
                changes: {
                    from: target.pos,
                    to: target.pos + target.text.length,
                    insert: text,
                },
            });
        };
    },
    setTarget(target) {
        this.target = target;
        view.dispatch({ effects: setInteract.of(target) });
    },
    isModKeyDown(e) {
        const modkey = view.state.facet(interactModKey);
        const isMac = Boolean(window.navigator) &&
            window.navigator.userAgent.includes('Macintosh');
        switch (modkey) {
            case "alt": return e.altKey;
            case "shift": return e.shiftKey;
            case "ctrl": return e.ctrlKey;
            case "meta": return e.metaKey;
            case "mod": return isMac ? e.metaKey : e.ctrlKey;
        }
        throw new Error(`Invalid mod key: ${modkey}`);
    },
    update(update) {
        const target = update.state.field(interactField, false);
        // the field isn't mounted
        if (target === undefined) {
            return;
        }
        if (this.target !== target) {
            this.target = target;
            if (target === null) {
                this.dragging = false;
            }
        }
    },
}), {
    eventHandlers: {
        mousedown(e, _view) {
            if (!this.isModKeyDown(e))
                return;
            const match = this.getMatch();
            if (!match)
                return;
            e.preventDefault();
            if (match.rule.onClick) {
                match.rule.onClick(match.text, this.updateText(match), e);
            }
            if (match.rule.onDrag) {
                this.dragging = true;
            }
        },
        mousemove(e, _view) {
            this.mouseX = e.clientX;
            this.mouseY = e.clientY;
            if (!this.isModKeyDown(e)) {
                if (this.target) {
                    this.setTarget(null);
                }
                return;
            }
            if (this.target && this.dragging) {
                if (this.target.rule.onDrag) {
                    this.target.rule.onDrag(this.target.text, this.updateText(this.target), e);
                }
            }
            else {
                this.setTarget(this.getMatch());
            }
        },
        mouseup(e, _view) {
            this.dragging = false;
            if (this.target && !this.isModKeyDown(e)) {
                this.setTarget(null);
            }
            if (this.isModKeyDown(e)) {
                this.setTarget(this.getMatch());
            }
        },
        mouseleave(_e, _view) {
            this.dragging = false;
            if (this.target) {
                this.setTarget(null);
            }
        },
        // TODO: fix these keybindings
        // these currently don't do anything because CodeMirror's keybinding
        // system prevents these events from firing.
        keydown(e, _view) {
            if (!this.target && this.isModKeyDown(e)) {
                this.setTarget(this.getMatch());
            }
        },
        keyup(e, _view) {
            if (this.target && !this.isModKeyDown(e)) {
                this.setTarget(null);
            }
        },
    },
});
const interact = (cfg = {}) => {
    var _a, _b;
    return [
        interactField,
        interactTheme,
        interactViewPlugin,
        interactModKey.of((_a = cfg.key) !== null && _a !== void 0 ? _a : "alt"),
        ((_b = cfg.rules) !== null && _b !== void 0 ? _b : []).map((r) => interactRule.of(r)),
    ];
};

exports["default"] = interact;
exports.interactModKey = interactModKey;
exports.interactRule = interactRule;
